#define PYGGI_MODULE
#include "pyggi.h"

staticforward PyTypeObject visual_type;
staticforward PyTypeObject pack_type;
staticforward PyTypeObject mode_type;

static Pygii_C_api * gii_c_api;

/* helpers */
static inline void float2color(double r, double g, double b, ggi_color *c)  {
	c->r=(int)(r*65535);
	c->g=(int)(g*65535);
	c->b=(int)(b*65535);
}

static inline void color2float(ggi_color*c, double *r, double *g,double *b)  {
	*r=((double)c->r)/65535;
	*g=((double)c->g)/65535;
	*b=((double)c->b)/65535;
}

static inline int pyggi_get_int(PyObject *o, int *i) {
	int a;
	a=PyInt_AsLong(o);
	if(PyErr_Occurred())
		return -1;
	*i=a;
	return 0;
}

static inline int pyggi_get_coords(PyObject *o, ggi_coord *c) {
	int r[2];
	if(PySequence_Check(o)) {
		if(PySequence_Size(o)==2) {
			r[0]=PyInt_AsLong(PySequence_GetItem(o,0));
			r[1]=PyInt_AsLong(PySequence_GetItem(o,1));
			if(PyErr_Occurred())
				return -1;
			c->x=r[0];
			c->y=r[1];
			return 0;
		}
	}
	PyErr_SetString(PyExc_TypeError,"a 2-int sequence is required");
	return -1;
}

static inline int pyggi_get_rect(PyObject *o, int i[4]) {
	int r[4];
	if(PySequence_Check(o)) {
		if(PySequence_Size(o)==4) {
			r[0]=PyInt_AsLong(PySequence_GetItem(o,0));
			r[1]=PyInt_AsLong(PySequence_GetItem(o,1));
			r[2]=PyInt_AsLong(PySequence_GetItem(o,2));
			r[3]=PyInt_AsLong(PySequence_GetItem(o,3));	     
			if(PyErr_Occurred())
				return -1;
			i[0]=r[0];
			i[1]=r[1];
			i[2]=r[2];
			i[3]=r[3];	
			return 0;
		}
	}
	
	PyErr_SetString(PyExc_TypeError,"a 4-int sequence is required");
	return -1;
}

static inline int pyggi_get_pixel(PyObject *o, ggi_pixel *p, ggi_visual * v) {
	ggi_color c;
	float r,g,b;
	if(PyInt_Check(o)) {
		*p=PyInt_AsLong(o);
		return 0;
	}
	if(PySequence_Check(o)) {
		if(PySequence_Size(o)==3) {
			r=PyFloat_AsDouble(PySequence_GetItem(o,0));
			g=PyFloat_AsDouble(PySequence_GetItem(o,1));
			b=PyFloat_AsDouble(PySequence_GetItem(o,2));
			if(PyErr_Occurred())
				return -1;
			float2color(r,g,b,&c);
			*p=ggiMapColor(v,&c);
			return 0;
		}
	}
	PyErr_SetString(PyExc_ValueError,"an int or a 3-float sequence is required");
	return -1;
}

/*********************************************************
 * Extension handling functions                          *
 *********************************************************/
static int pyggi_extension_attach(GGIVisualObject  *vis,
				  char          *n,
				  ExtGetAttrFunc ga,
				  ExtSetAttrFunc sa) {
	GGIExtension * e;
	
	/* test if it's not already attached */
	
	for(e=vis->exts;e;e=e->next)
		if(!strcmp(e->name,n))
			return -1;
	e=malloc(sizeof(GGIExtension));
	if(e==NULL)
		return -2;
	
	e->next=vis->exts;
	e->name=n;
	e->getattr=ga;
	e->setattr=sa;
	vis->exts=e;
	vis->nb_exts++;
	return 0;
}

static int pyggi_extension_detach(GGIVisualObject  *vis,
				  char          *name) {
	GGIExtension ** t;
	GGIExtension * e;
	
	for(e=vis->exts,t=&vis->exts;e;e=e->next,t=&e->next) {
		if(!strcmp(e->name,name)) {
			/* Have to remove this one*/
			*t=e->next;			
			free(e);
			vis->nb_exts--;
			return 0;
		}
	}

	/* Not registered */

	return -1;
}

static void pyggi_extension_detach_all(GGIVisualObject  *vis) {
	GGIExtension * r;
	
	while(vis->exts) {
		r=vis->exts->next;
		free(vis->exts);
		vis->exts=r;
	}
	vis->nb_exts=0;
}


static PyObject * pyggi_extension_list(GGIVisualObject  *vis) {
	PyObject * r;
	GGIExtension * e;
	int i;
	
	r=PyList_New(vis->nb_exts);
	if(r) {
		e=vis->exts;
		for(i=0;i<vis->nb_exts;i++,e=e->next) {
			PyList_SET_ITEM(r,i,PyString_FromString(e->name));
		}
	}
	return r;
}

#include "pyggi_mode.inc"
#include "pyggi_visual.inc"

/**************************************************************/
/* Module functions                                           */
/**************************************************************/
const char ggi_init_doc[]="
Initialize GGI. It must be called at least one before using any other function.
Successive calls to this function has no effect but increasing the use count.

Parameters: None
Return:     None
";
static PyObject* ggi_init(PyObject*self,PyObject*args) {
	if(ggiInit()!=0) {
		PyErr_SetString(PyExc_SystemError, "ggiInit() failed");
		return NULL;
	}
	Py_INCREF(Py_None); return Py_None;
}

const char ggi_exit_doc[]="
Terminate a GGI session. GGI is really terminated when the use count is down to 0.

Parameters: None
Return:     None
";
static PyObject* ggi_exit(PyObject*self,PyObject*args) {
	if (ggiExit()<0) {
		PyErr_SetString(PyExc_SystemError, "ggiExit() failed");
		return NULL;		
	}
	Py_INCREF(Py_None); return Py_None;
}

const char ggi_open_doc[]="
Open the display given as argument (a single string). If no argument is given, it
tries to open the default display as set in the GGI_DISPLAY environment variable.

Parameters: ()|(s) : An optional string describing the display
Return:     O : A new Visual object
";
static PyObject* ggi_open(PyObject*self, PyObject*args) {
	return newVisualObject(args);
}

static struct PyMethodDef ggi_methods[] = {
	{"ggiInit",  (PyCFunction)ggi_init,      METH_NOARGS,  (char*)ggi_init_doc },
	{"ggiExit",  (PyCFunction)ggi_exit,      METH_NOARGS,  (char*)ggi_exit_doc },
	{"Visual",   (PyCFunction)ggi_open,      METH_VARARGS, (char*)ggi_open_doc },
	{"Mode",     (PyCFunction)ModeConstruct, METH_VARARGS, (char*)mode_construct_doc},
	{NULL, NULL}
};

const char ggi_doc[]="
PyGGI 0.1

PyGGI allows python programmers to access the GGI display system through an
elegant and effective (in the author opinion) python interface.

The first thing to do is to call ggiInit().
Good luck.

Author
- Eric Faurot <eric.faurot@info.unicaen.fr>
";

void initggi(void) {
	PyObject *m,*d,*t;
	PyObject *gii;
	PyObject *giidict;
	PyObject *capiobj;
	static Pyggi_C_api capi;
	
	m=Py_InitModule4("ggi", ggi_methods, (char *)ggi_doc,
			 (PyObject*)NULL, PYTHON_API_VERSION);
	
	d=PyModule_GetDict(m);
	
	/* export C API */
	capi.visual_type = &visual_type;
	capi.mode_type   = &mode_type;

	capi.attach = pyggi_extension_attach;
	capi.detach = pyggi_extension_detach;

	t=PyCObject_FromVoidPtr(&capi, NULL);
        PyDict_SetItemString(d, "_ggi_C_API", t);
        Py_DECREF(t);

	/* import gii */
	
	gii= PyImport_ImportModule("PyGGI.gii");
	if (gii!= NULL) {
		giidict=PyModule_GetDict(gii);
		capiobj=PyDict_GetItemString(giidict, "_gii_C_API");
		if (PyCObject_Check(capiobj)) {
			gii_c_api=(Pygii_C_api *)PyCObject_AsVoidPtr(capiobj);
		}
	}

	/* set the dictonary */
	PyDict_SetItemString(d, "flagAsync", Py_BuildValue("i", GGIFLAG_ASYNC));

	if(PyErr_Occurred()) Py_FatalError("Can't initialize module ggiwrapper");
}
